Block-Structured AMR Software Framework
Loading...
Searching...
No Matches
AMReX_ParticleIO.H
Go to the documentation of this file.
1#ifndef AMREX_PARTICLEIO_H
2#define AMREX_PARTICLEIO_H
3#include <AMReX_Config.H>
4
7#include <AMReX_VisMF.H>
8
9#include <iterator>
10
11namespace amrex {
12
13template <typename ParticleType, int NArrayReal, int NArrayInt,
14 template<class> class Allocator, class CellAssignor>
15void
16ParticleContainer_impl<ParticleType, NArrayReal, NArrayInt, Allocator, CellAssignor>
17::WriteParticleRealData (void* data, size_t size, std::ostream& os) const
18{
19 if (sizeof(typename ParticleType::RealType) == 4) {
20 writeFloatData((float*) data, size, os, ParticleRealDescriptor);
21 }
22 else if (sizeof(typename ParticleType::RealType) == 8) {
23 writeDoubleData((double*) data, size, os, ParticleRealDescriptor);
24 }
25}
26
27template <typename ParticleType, int NArrayReal, int NArrayInt,
28 template<class> class Allocator, class CellAssignor>
29void
31::ReadParticleRealData (void* data, size_t size, std::istream& is)
32{
33 if (sizeof(typename ParticleType::RealType) == 4) {
34 readFloatData((float*) data, size, is, ParticleRealDescriptor);
35 }
36 else if (sizeof(typename ParticleType::RealType) == 8) {
37 readDoubleData((double*) data, size, is, ParticleRealDescriptor);
38 }
39}
40
42 template <typename P>
44 int operator() (const P& p) const {
45 return p.id().is_valid();
46 }
47};
48
50template <typename ParticleType, int NArrayReal, int NArrayInt,
51 template<class> class Allocator, class CellAssignor>
52void
53ParticleContainer_impl<ParticleType, NArrayReal, NArrayInt, Allocator, CellAssignor>
54::Checkpoint (const std::string& dir,
55 const std::string& name, bool /*is_checkpoint*/,
56 const Vector<std::string>& real_comp_names,
57 const Vector<std::string>& int_comp_names) const
58{
59 Vector<int> write_real_comp;
60 Vector<std::string> tmp_real_comp_names;
61
62 int first_rcomp = ParticleType::is_soa_particle ? AMREX_SPACEDIM : 0;
63 for (int i = first_rcomp; i < NStructReal + NumRealComps(); ++i )
64 {
65 write_real_comp.push_back(1);
66 if (real_comp_names.empty())
67 {
68 tmp_real_comp_names.push_back(getDefaultCompNameReal<ParticleType>(i));
69 }
70 else
71 {
72 tmp_real_comp_names.push_back(real_comp_names[i-first_rcomp]);
73 }
74 }
75
76 Vector<int> write_int_comp;
77 Vector<std::string> tmp_int_comp_names;
78 for (int i = 0; i < NStructInt + NumIntComps(); ++i )
79 {
80 write_int_comp.push_back(1);
81 if (int_comp_names.empty())
82 {
83 tmp_int_comp_names.push_back(getDefaultCompNameInt<ParticleType>(i));
84 }
85 else
86 {
87 tmp_int_comp_names.push_back(int_comp_names[i]);
88 }
89 }
90
91 WriteBinaryParticleData(dir, name, write_real_comp, write_int_comp,
92 tmp_real_comp_names, tmp_int_comp_names,
93 FilterPositiveID{}, true);
94}
95
96template <typename ParticleType, int NArrayReal, int NArrayInt,
97 template<class> class Allocator, class CellAssignor>
98void
100::Checkpoint (const std::string& dir, const std::string& name,
101 const Vector<int>& write_real_comp,
102 const Vector<int>& write_int_comp,
103 const Vector<std::string>& real_comp_names,
104 const Vector<std::string>& int_comp_names) const
105{
106 WriteBinaryParticleData(dir, name, write_real_comp, write_int_comp,
107 real_comp_names, int_comp_names,
108 FilterPositiveID{}, true);
109}
110
112template <typename ParticleType, int NArrayReal, int NArrayInt,
113 template<class> class Allocator, class CellAssignor>
114void
116::WritePlotFile (const std::string& dir, const std::string& name) const
117{
118 Vector<int> write_real_comp;
119 Vector<std::string> real_comp_names;
120
121 int first_rcomp = ParticleType::is_soa_particle ? AMREX_SPACEDIM : 0;
122 for (int i = first_rcomp; i < NStructReal + NumRealComps(); ++i )
123 {
124 write_real_comp.push_back(1);
125 real_comp_names.push_back(getDefaultCompNameReal<ParticleType>(i));
126 }
127
128 Vector<int> write_int_comp;
129 Vector<std::string> int_comp_names;
130 for (int i = 0; i < NStructInt + NumIntComps(); ++i )
131 {
132 write_int_comp.push_back(1);
133 int_comp_names.push_back(getDefaultCompNameInt<ParticleType>(i));
134 }
135
136 WriteBinaryParticleData(dir, name, write_real_comp, write_int_comp,
137 real_comp_names, int_comp_names,
139}
140
142template <typename ParticleType, int NArrayReal, int NArrayInt,
143 template<class> class Allocator, class CellAssignor>
144void
146::WritePlotFile (const std::string& dir, const std::string& name,
147 const Vector<std::string>& real_comp_names,
148 const Vector<std::string>& int_comp_names) const
149{
150 if constexpr(ParticleType::is_soa_particle) {
151 AMREX_ALWAYS_ASSERT(real_comp_names.size() == NumRealComps() + NStructReal - AMREX_SPACEDIM); // pure SoA: skip positions
152 } else {
153 AMREX_ALWAYS_ASSERT(real_comp_names.size() == NumRealComps() + NStructReal);
154 }
155 AMREX_ASSERT( int_comp_names.size() == NStructInt + NumIntComps() );
156
157 Vector<int> write_real_comp;
158 int nrc = ParticleType::is_soa_particle ? NStructReal + NumRealComps() - AMREX_SPACEDIM : NStructReal + NumRealComps();
159 for (int i = 0; i < nrc; ++i) {
160 write_real_comp.push_back(1);
161 }
162
163 Vector<int> write_int_comp;
164 for (int i = 0; i < NStructInt + NumIntComps(); ++i) {
165 write_int_comp.push_back(1);
166 }
167
168 WriteBinaryParticleData(dir, name,
169 write_real_comp, write_int_comp,
170 real_comp_names, int_comp_names,
172}
173
175template <typename ParticleType, int NArrayReal, int NArrayInt,
176 template<class> class Allocator, class CellAssignor>
177void
179::WritePlotFile (const std::string& dir, const std::string& name,
180 const Vector<std::string>& real_comp_names) const
181{
182 if constexpr(ParticleType::is_soa_particle) {
183 AMREX_ALWAYS_ASSERT(real_comp_names.size() == NumRealComps() + NStructReal - AMREX_SPACEDIM); // pure SoA: skip positions
184 } else {
185 AMREX_ALWAYS_ASSERT(real_comp_names.size() == NumRealComps() + NStructReal);
186 }
187
188 Vector<int> write_real_comp;
189 int nrc = ParticleType::is_soa_particle ? NStructReal + NumRealComps() - AMREX_SPACEDIM : NStructReal + NumRealComps();
190 for (int i = 0; i < nrc; ++i) {
191 write_real_comp.push_back(1);
192 }
193
194 Vector<int> write_int_comp;
195 for (int i = 0; i < NStructInt + NumIntComps(); ++i) {
196 write_int_comp.push_back(1);
197 }
198
199 Vector<std::string> int_comp_names;
200 for (int i = 0; i < NStructInt + NumIntComps(); ++i )
201 {
202 int_comp_names.push_back(getDefaultCompNameInt<ParticleType>(i));
203 }
204
205 WriteBinaryParticleData(dir, name,
206 write_real_comp, write_int_comp,
207 real_comp_names, int_comp_names,
209}
210
212template <typename ParticleType, int NArrayReal, int NArrayInt,
213 template<class> class Allocator, class CellAssignor>
214void
216::WritePlotFile (const std::string& dir,
217 const std::string& name,
218 const Vector<int>& write_real_comp,
219 const Vector<int>& write_int_comp) const
220{
221
222 if constexpr(ParticleType::is_soa_particle) {
223 AMREX_ALWAYS_ASSERT(write_real_comp.size() == NumRealComps() + NStructReal - AMREX_SPACEDIM); // pure SoA: skip positions
224 } else {
225 AMREX_ALWAYS_ASSERT(write_real_comp.size() == NumRealComps() + NStructReal);
226 }
227 AMREX_ASSERT(write_int_comp.size() == NStructInt + NumIntComps() );
228
229 Vector<std::string> real_comp_names;
230 int first_rcomp = ParticleType::is_soa_particle ? AMREX_SPACEDIM : 0;
231 for (int i = first_rcomp; i < NStructReal + NumRealComps(); ++i )
232 {
233 real_comp_names.push_back(getDefaultCompNameReal<ParticleType>(i));
234 }
235
236 Vector<std::string> int_comp_names;
237 for (int i = 0; i < NStructInt + NumIntComps(); ++i )
238 {
239 int_comp_names.push_back(getDefaultCompNameInt<ParticleType>(i));
240 }
241
242 WriteBinaryParticleData(dir, name, write_real_comp, write_int_comp,
243 real_comp_names, int_comp_names,
245}
246
248template <typename ParticleType, int NArrayReal, int NArrayInt,
249 template<class> class Allocator, class CellAssignor>
250void
252WritePlotFile (const std::string& dir, const std::string& name,
253 const Vector<int>& write_real_comp,
254 const Vector<int>& write_int_comp,
255 const Vector<std::string>& real_comp_names,
256 const Vector<std::string>& int_comp_names) const
257{
258 BL_PROFILE("ParticleContainer::WritePlotFile()");
259
260 WriteBinaryParticleData(dir, name,
261 write_real_comp, write_int_comp,
262 real_comp_names, int_comp_names,
264}
265
267template <typename ParticleType, int NArrayReal, int NArrayInt,
268 template<class> class Allocator, class CellAssignor>
269template <class F>
270void
272::WritePlotFile (const std::string& dir, const std::string& name, F const& f) const
273{
274 Vector<int> write_real_comp;
275 Vector<std::string> real_comp_names;
276
277 int first_rcomp = ParticleType::is_soa_particle ? AMREX_SPACEDIM : 0;
278 for (int i = first_rcomp; i < NStructReal + NumRealComps(); ++i )
279 {
280 write_real_comp.push_back(1);
281 real_comp_names.push_back(getDefaultCompNameReal<ParticleType>(i));
282 }
283
284 Vector<int> write_int_comp;
285 Vector<std::string> int_comp_names;
286 for (int i = 0; i < NStructInt + NumIntComps(); ++i )
287 {
288 write_int_comp.push_back(1);
289 int_comp_names.push_back(getDefaultCompNameInt<ParticleType>(i));
290 }
291
292 WriteBinaryParticleData(dir, name, write_real_comp, write_int_comp,
293 real_comp_names, int_comp_names, f);
294}
295
297template <typename ParticleType, int NArrayReal, int NArrayInt,
298 template<class> class Allocator, class CellAssignor>
299template <class F>
300void
302::WritePlotFile (const std::string& dir, const std::string& name,
303 const Vector<std::string>& real_comp_names,
304 const Vector<std::string>& int_comp_names, F const& f) const
305{
306 if constexpr(ParticleType::is_soa_particle) {
307 AMREX_ALWAYS_ASSERT(real_comp_names.size() == NumRealComps() + NStructReal - AMREX_SPACEDIM); // pure SoA: skip positions
308 } else {
309 AMREX_ALWAYS_ASSERT(real_comp_names.size() == NumRealComps() + NStructReal);
310 }
311 AMREX_ASSERT( int_comp_names.size() == NStructInt + NumIntComps() );
312
313 Vector<int> write_real_comp;
314 int nrc = ParticleType::is_soa_particle ? NStructReal + NumRealComps() - AMREX_SPACEDIM : NStructReal + NumRealComps();
315 for (int i = 0; i < nrc; ++i) {
316 write_real_comp.push_back(1);
317 }
318
319 Vector<int> write_int_comp;
320 for (int i = 0; i < NStructInt + NumIntComps(); ++i) {
321 write_int_comp.push_back(1);
322 }
323
324 WriteBinaryParticleData(dir, name,
325 write_real_comp, write_int_comp,
326 real_comp_names, int_comp_names, f);
327}
328
330template <typename ParticleType, int NArrayReal, int NArrayInt,
331 template<class> class Allocator, class CellAssignor>
332template <class F>
333void
335::WritePlotFile (const std::string& dir, const std::string& name,
336 const Vector<std::string>& real_comp_names, F const& f) const
337{
338 if constexpr(ParticleType::is_soa_particle) {
339 AMREX_ALWAYS_ASSERT(real_comp_names.size() == NumRealComps() + NStructReal - AMREX_SPACEDIM); // pure SoA: skip positions
340 } else {
341 AMREX_ALWAYS_ASSERT(real_comp_names.size() == NumRealComps() + NStructReal);
342 }
343
344 Vector<int> write_real_comp;
345 int nrc = ParticleType::is_soa_particle ? NStructReal + NumRealComps() - AMREX_SPACEDIM : NStructReal + NumRealComps();
346 for (int i = 0; i < nrc; ++i) {
347 write_real_comp.push_back(1);
348 }
349
350 Vector<int> write_int_comp;
351 for (int i = 0; i < NStructInt + NumIntComps(); ++i) {
352 write_int_comp.push_back(1);
353 }
354
355 Vector<std::string> int_comp_names;
356 for (int i = 0; i < NStructInt + NumIntComps(); ++i )
357 {
358 int_comp_names.push_back(getDefaultCompNameInt<ParticleType>(i));
359 }
360
361 WriteBinaryParticleData(dir, name,
362 write_real_comp, write_int_comp,
363 real_comp_names, int_comp_names, f);
364}
365
367template <typename ParticleType, int NArrayReal, int NArrayInt,
368 template<class> class Allocator, class CellAssignor>
369template <class F>
370void
372::WritePlotFile (const std::string& dir,
373 const std::string& name,
374 const Vector<int>& write_real_comp,
375 const Vector<int>& write_int_comp, F const& f) const
376{
377 if constexpr(ParticleType::is_soa_particle) {
378 AMREX_ALWAYS_ASSERT(write_real_comp.size() == NumRealComps() + NStructReal - AMREX_SPACEDIM); // pure SoA: skip positions
379 } else {
380 AMREX_ALWAYS_ASSERT(write_real_comp.size() == NumRealComps() + NStructReal);
381 }
382 AMREX_ASSERT(write_int_comp.size() == NStructInt + NumIntComps() );
383
384 Vector<std::string> real_comp_names;
385 int first_rcomp = ParticleType::is_soa_particle ? AMREX_SPACEDIM : 0;
386 for (int i = first_rcomp; i < NStructReal + NumRealComps(); ++i )
387 {
388 real_comp_names.push_back(getDefaultCompNameReal<ParticleType>(i));
389 }
390
391 Vector<std::string> int_comp_names;
392 for (int i = 0; i < NStructInt + NumIntComps(); ++i )
393 {
394 int_comp_names.push_back(getDefaultCompNameInt<ParticleType>(i));
395 }
396
397 WriteBinaryParticleData(dir, name, write_real_comp, write_int_comp,
398 real_comp_names, int_comp_names, f);
399}
400
402template <typename ParticleType, int NArrayReal, int NArrayInt,
403 template<class> class Allocator, class CellAssignor>
404template <class F>
405void
407WritePlotFile (const std::string& dir, const std::string& name,
408 const Vector<int>& write_real_comp,
409 const Vector<int>& write_int_comp,
410 const Vector<std::string>& real_comp_names,
411 const Vector<std::string>& int_comp_names,
412 F const& f) const
413{
414 BL_PROFILE("ParticleContainer::WritePlotFile()");
415
416 WriteBinaryParticleData(dir, name,
417 write_real_comp, write_int_comp,
418 real_comp_names, int_comp_names, f);
419}
420
421template <typename ParticleType, int NArrayReal, int NArrayInt,
422 template<class> class Allocator, class CellAssignor>
423template <class F>
424void
426::WriteBinaryParticleData (const std::string& dir, const std::string& name,
427 const Vector<int>& write_real_comp,
428 const Vector<int>& write_int_comp,
429 const Vector<std::string>& real_comp_names,
430 const Vector<std::string>& int_comp_names,
431 F const& f, bool is_checkpoint) const
432{
433 if (AsyncOut::UseAsyncOut()) {
434 WriteBinaryParticleDataAsync(*this, dir, name,
435 write_real_comp, write_int_comp,
436 real_comp_names, int_comp_names, is_checkpoint);
437 } else
438 {
439 WriteBinaryParticleDataSync(*this, dir, name,
440 write_real_comp, write_int_comp,
441 real_comp_names, int_comp_names,
442 f, is_checkpoint);
443 }
444}
445
446template <typename ParticleType, int NArrayReal, int NArrayInt,
447 template<class> class Allocator, class CellAssignor>
448void
451{
452 if( ! usePrePost) {
453 return;
454 }
455
456 BL_PROFILE("ParticleContainer::CheckpointPre()");
457
458 const int IOProcNumber = ParallelDescriptor::IOProcessorNumber();
459 Long nparticles = 0;
460 Long maxnextid = ParticleType::NextID();
461
462 for (int lev = 0; lev < std::ssize(m_particles); lev++) {
463 const auto& pmap = m_particles[lev];
464 for (const auto& kv : pmap) {
465 const auto& aos = kv.second.GetArrayOfStructs();
466 for (int k = 0; k < aos.numParticles(); ++k) {
467 const ParticleType& p = aos[k];
468 if (p.id().is_valid()) {
469 //
470 // Only count (and checkpoint) valid particles.
471 //
472 nparticles++;
473 }
474 }
475 }
476 }
477 ParallelDescriptor::ReduceLongSum(nparticles, IOProcNumber);
478
479 ParticleType::NextID(maxnextid);
480 ParallelDescriptor::ReduceLongMax(maxnextid, IOProcNumber);
481
482 nparticlesPrePost = nparticles;
483 maxnextidPrePost = maxnextid;
484
485 nParticlesAtLevelPrePost.clear();
486 nParticlesAtLevelPrePost.resize(finestLevel() + 1, 0);
487 for(int lev(0); lev <= finestLevel(); ++lev) {
488 nParticlesAtLevelPrePost[lev] = NumberOfParticlesAtLevel(lev);
489 }
490
491 whichPrePost.clear();
492 whichPrePost.resize(finestLevel() + 1);
493 countPrePost.clear();
494 countPrePost.resize(finestLevel() + 1);
495 wherePrePost.clear();
496 wherePrePost.resize(finestLevel() + 1);
497
498 filePrefixPrePost.clear();
499 filePrefixPrePost.resize(finestLevel() + 1);
500}
501
502
503template <typename ParticleType, int NArrayReal, int NArrayInt,
504 template<class> class Allocator, class CellAssignor>
505void
508{
509 if( ! usePrePost) {
510 return;
511 }
512
513 BL_PROFILE("ParticleContainer::CheckpointPost()");
514
515 const int IOProcNumber = ParallelDescriptor::IOProcessorNumber();
516 std::ofstream HdrFile;
517 HdrFile.open(HdrFileNamePrePost.c_str(), std::ios::out | std::ios::app);
518
519 for(int lev(0); lev <= finestLevel(); ++lev) {
520 ParallelDescriptor::ReduceIntSum (whichPrePost[lev].dataPtr(), whichPrePost[lev].size(), IOProcNumber);
521 ParallelDescriptor::ReduceIntSum (countPrePost[lev].dataPtr(), countPrePost[lev].size(), IOProcNumber);
522 ParallelDescriptor::ReduceLongSum(wherePrePost[lev].dataPtr(), wherePrePost[lev].size(), IOProcNumber);
523
524
526 for(int j(0); j < whichPrePost[lev].size(); ++j) {
527 HdrFile << whichPrePost[lev][j] << ' ' << countPrePost[lev][j] << ' ' << wherePrePost[lev][j] << '\n';
528 }
529
530 const bool gotsome = (nParticlesAtLevelPrePost[lev] > 0);
531 if(gotsome && doUnlink) {
532// BL_PROFILE_VAR("PC<NNNN>::Checkpoint:unlink", unlink_post);
533 // Unlink any zero-length data files.
534 Vector<Long> cnt(nOutFilesPrePost,0);
535
536 for(int i(0), N = countPrePost[lev].size(); i < N; ++i) {
537 cnt[whichPrePost[lev][i]] += countPrePost[lev][i];
538 }
539
540 for(int i = 0; i < std::ssize(cnt); ++i) {
541 if(cnt[i] == 0) {
542 std::string FullFileName = NFilesIter::FileName(i, filePrefixPrePost[lev]);
543 FileSystem::Remove(FullFileName);
544 }
545 }
546 }
547 }
548 }
549
551 HdrFile.flush();
552 HdrFile.close();
553 if( ! HdrFile.good()) {
554 amrex::Abort("ParticleContainer::CheckpointPost(): problem writing HdrFile");
555 }
556 }
557}
558
559template <typename ParticleType, int NArrayReal, int NArrayInt,
560 template<class> class Allocator, class CellAssignor>
561void
564{
565 CheckpointPre();
566}
567
568
569template <typename ParticleType, int NArrayReal, int NArrayInt,
570 template<class> class Allocator, class CellAssignor>
571void
574{
575 CheckpointPost();
576}
577
578
579template <typename ParticleType, int NArrayReal, int NArrayInt,
580 template<class> class Allocator, class CellAssignor>
581void
583::WriteParticles (int lev, std::ofstream& ofs, int fnum,
584 Vector<int>& which, Vector<int>& count, Vector<Long>& where,
585 const Vector<int>& write_real_comp,
586 const Vector<int>& write_int_comp,
587 const Vector<std::map<std::pair<int, int>, FlagsVector>>& particle_io_flags,
588 bool is_checkpoint) const
589{
590 BL_PROFILE("ParticleContainer::WriteParticles()");
591
592 // For a each grid, the tiles it contains
593 std::map<int, Vector<int> > tile_map;
594
595 for (const auto& kv : m_particles[lev])
596 {
597 const int grid = kv.first.first;
598 const int tile = kv.first.second;
599 tile_map[grid].push_back(tile);
600 const auto& pflags = particle_io_flags[lev].at(kv.first);
601
602 // Only write out valid particles.
603 count[grid] += particle_detail::countFlags(pflags);
604 }
605
606 MFInfo info;
607 info.SetAlloc(false);
608 MultiFab state(ParticleBoxArray(lev), ParticleDistributionMap(lev), 1,0,info);
609
610 for (MFIter mfi(state); mfi.isValid(); ++mfi)
611 {
612 const int grid = mfi.index();
613
614 which[grid] = fnum;
615 where[grid] = VisMF::FileOffset(ofs);
616
617 if (count[grid] <= 0) { continue; }
618
619 Vector<int> istuff;
621 particle_detail::packIOData(istuff, rstuff, *this, lev, grid,
622 write_real_comp, write_int_comp,
623 particle_io_flags, tile_map[grid], count[grid], is_checkpoint);
624
625 writeIntData(istuff.dataPtr(), istuff.size(), ofs);
627 ofs.flush(); // Some systems require this flush() (probably due to a bug)
628 }
629
630 WriteParticleRealData(rstuff.dataPtr(), rstuff.size(), ofs);
632 ofs.flush(); // Some systems require this flush() (probably due to a bug)
633 }
634 }
635}
636
637
638template <typename ParticleType, int NArrayReal, int NArrayInt,
639 template<class> class Allocator, class CellAssignor>
640void
642::Restart (const std::string& dir, const std::string& file, bool /*is_checkpoint*/)
643{
644 Restart(dir, file);
645}
646
647template <typename ParticleType, int NArrayReal, int NArrayInt,
648 template<class> class Allocator, class CellAssignor>
649void
651::Restart (const std::string& dir, const std::string& file)
652{
653 BL_PROFILE("ParticleContainer::Restart()");
654 AMREX_ASSERT(!dir.empty());
655 AMREX_ASSERT(!file.empty());
656
657 const auto strttime = amrex::second();
658
659 int DATA_Digits_Read(5);
660 ParmParse pp("particles");
661 pp.query("datadigits_read",DATA_Digits_Read);
662
663 std::string fullname = dir;
664 if (!fullname.empty() && fullname[fullname.size()-1] != '/') {
665 fullname += '/';
666 }
667 fullname += file;
668 std::string HdrFileName = fullname;
669 if (!HdrFileName.empty() && HdrFileName[HdrFileName.size()-1] != '/') {
670 HdrFileName += '/';
671 }
672 HdrFileName += "Header";
673
674 Vector<char> fileCharPtr;
675 ParallelDescriptor::ReadAndBcastFile(HdrFileName, fileCharPtr);
676 std::string fileCharPtrString(fileCharPtr.dataPtr());
677 std::istringstream HdrFile(fileCharPtrString, std::istringstream::in);
678
679 // The metadata prefix (version through finest level) and the per-level
680 // grid table (data file index, particle count and byte offset per grid)
681 // are parsed by the shared ParticleHeader.
682 // See AMReX_ParticleHeader.{H,cpp}.
683 ParticleHeader header;
684 header.parse(HdrFile);
685 header.parse_grid_table(HdrFile);
686
687 const std::string& how = header.how;
688 const bool convert_ids = header.convert_ids;
689
690 if (header.dim != AMREX_SPACEDIM) {
691 amrex::Abort("ParticleContainer::Restart(): dm != AMREX_SPACEDIM");
692 }
693
694 const int nrc = ParticleType::is_soa_particle ? NStructReal + NumRealComps() - AMREX_SPACEDIM : NStructReal + NumRealComps();
695 if (header.num_real != nrc) {
696 amrex::Abort("ParticleContainer::Restart(): nr not the expected value");
697 }
698
699 if (header.num_int != NStructInt + NumIntComps()) {
700 amrex::Abort("ParticleContainer::Restart(): ni != NStructInt");
701 }
702
703 AMREX_ASSERT(header.num_particles >= 0);
704
705 AMREX_ASSERT(header.next_id > 0);
706 ParticleType::NextID(header.next_id);
707
708 const int finest_level_in_file = header.finest_level;
709 AMREX_ASSERT(finest_level_in_file >= 0);
710
711 // Read the BoxArrays stored in the checkpoint for each level so we can
712 // drive the read using the file's layout. The container's own grids are
713 // never modified: Redistribute() will move every particle to the correct
714 // tile on whatever grids the container currently uses.
715 Vector<BoxArray> file_bas(finest_level_in_file + 1);
716 Vector<DistributionMapping> file_dms(finest_level_in_file + 1);
717 for (int lev = 0; lev <= finest_level_in_file; lev++)
718 {
719 std::string phdr_name = fullname;
720 phdr_name += "/Level_";
721 phdr_name = amrex::Concatenate(phdr_name, lev, 1);
722 phdr_name += "/Particle_H";
723
724 if (amrex::FileExists(phdr_name)) {
725 Vector<char> phdr_chars;
726 ParallelDescriptor::ReadAndBcastFile(phdr_name, phdr_chars);
727 std::string phdr_string(phdr_chars.dataPtr());
728 std::istringstream phdr_file(phdr_string, std::istringstream::in);
729 file_bas[lev].readFrom(phdr_file);
730 } else {
731 // No particles were written at this level; use a single-box BA as
732 // a placeholder so that MFIter below works uniformly.
733 file_bas[lev] = BoxArray(Geom(lev).Domain());
734 }
735 const int NReaders = MaxReaders();
736 if (ParallelDescriptor::NProcs() <= NReaders) {
737 file_dms[lev] = DistributionMapping(file_bas[lev]);
738 } else {
739 Vector<int> pmap(file_bas[lev].size());
740 for (int i = 0; i < (int)pmap.size(); ++i) { pmap[i] = i % NReaders; }
741 file_dms[lev] = DistributionMapping(pmap);
742 }
743 }
744
745 resizeData();
746
747 if (finest_level_in_file > finestLevel()) {
748 m_particles.resize(finest_level_in_file+1);
749 }
750
751 for (int lev = 0; lev <= finest_level_in_file; lev++) {
752 auto const& grid_table = header.grids[lev];
753
754 // Use the file's own layout to distribute reading across ranks.
755 MFInfo info;
756 info.SetAlloc(false);
757 MultiFab file_mf(file_bas[lev], file_dms[lev], 1, 0, info);
758 Vector<int> grids_to_read;
759 for (MFIter mfi(file_mf); mfi.isValid(); ++mfi) {
760 grids_to_read.push_back(mfi.index());
761 }
762
763 for(int grid : grids_to_read) {
764 if (grid_table[grid].count <= 0) { continue; }
765
766 // The file names in the header file are relative.
767 std::string name = fullname;
769 if (!name.empty() && name[name.size()-1] != '/') {
770 name += '/';
771 }
772
773 name += "Level_";
774 name += amrex::Concatenate("", lev, 1);
775 name += '/';
776 name += DataPrefix();
777 name += amrex::Concatenate("", grid_table[grid].which, DATA_Digits_Read);
778
779 std::ifstream ParticleFile;
780
781 ParticleFile.open(name.c_str(), std::ios::in | std::ios::binary);
782
783 if (!ParticleFile.good()) {
785 }
786
787 ParticleFile.seekg(grid_table[grid].where, std::ios::beg);
788
789 // Use if constexpr to avoid instantiating the mis-matched
790 // type case and triggering the static_assert on the
791 // underlying copy calls
792 if (how == "single") {
793 if constexpr (std::is_same_v<ParticleReal, float>) {
794 ReadParticles<float>(grid_table[grid].count, grid, lev, ParticleFile, finest_level_in_file, convert_ids);
795 } else {
796 amrex::Error("File contains single-precision data, while AMReX is compiled with ParticleReal==double");
797 }
798 }
799 else if (how == "double") {
800 if constexpr (std::is_same_v<ParticleReal, double>) {
801 ReadParticles<double>(grid_table[grid].count, grid, lev, ParticleFile, finest_level_in_file, convert_ids);
802 } else {
803 amrex::Error("File contains double-precision data, while AMReX is compiled with ParticleReal==float");
804 }
805 }
806 else {
807 std::string msg("ParticleContainer::Restart(): bad parameter: ");
808 msg += how;
809 amrex::Error(msg.c_str());
810 }
811
812 ParticleFile.close();
813
814 if (!ParticleFile.good()) {
815 amrex::Abort("ParticleContainer::Restart(): problem reading particles");
816 }
817 }
818 }
819
820 Redistribute();
821
822 AMREX_ASSERT(OK());
823
824 if (m_verbose > 1) {
825 auto stoptime = amrex::second() - strttime;
827 amrex::Print() << "ParticleContainer::Restart() time: " << stoptime << '\n';
828 }
829}
830
831// Read a batch of particles from the checkpoint file
832template <typename ParticleType, int NArrayReal, int NArrayInt,
833 template<class> class Allocator, class CellAssignor>
834template <class RTYPE>
835void
837::ReadParticles (int cnt, int grd, int lev, std::ifstream& ifs,
838 int finest_level_in_file, bool convert_ids)
839{
840 BL_PROFILE("ParticleContainer::ReadParticles()");
841 AMREX_ASSERT(cnt > 0);
842 AMREX_ASSERT(lev < std::ssize(m_particles));
843
844 // First read in the integer data in binary. We do not store
845 // the m_lev and m_grid data on disk. We can easily recreate
846 // that given the structure of the checkpoint file.
847 const int iChunkSize = 2 + NStructInt + NumIntComps();
848 Vector<int> istuff(std::size_t(cnt)*iChunkSize);
849 readIntData(istuff.dataPtr(), istuff.size(), ifs, FPC::NativeIntDescriptor());
850
851 // Then the real data in binary.
852 const int rChunkSize = ParticleType::is_soa_particle ? NStructReal + NumRealComps() : AMREX_SPACEDIM + NStructReal + NumRealComps();
853 Vector<RTYPE> rstuff(std::size_t(cnt)*rChunkSize);
854 ReadParticleRealData(rstuff.dataPtr(), rstuff.size(), ifs);
855
856 // Now reassemble the particles.
857 int* iptr = istuff.dataPtr();
858 RTYPE* rptr = rstuff.dataPtr();
859
861 ParticleLocData pld;
862
864 host_particles.reserve(15);
865 host_particles.resize(finest_level_in_file+1);
866
868 std::vector<Gpu::HostVector<RTYPE> > > > host_real_attribs;
869 host_real_attribs.reserve(15);
870 host_real_attribs.resize(finest_level_in_file+1);
873 std::vector<Gpu::HostVector<int> > > > host_int_attribs;
874 host_int_attribs.reserve(15);
875 host_int_attribs.resize(finest_level_in_file+1);
876
878 host_idcpu.reserve(15);
879 host_idcpu.resize(finest_level_in_file+1);
880
881 for (int i = 0; i < cnt; i++) {
882 // note: for pure SoA particle layouts, we do write the id, cpu and positions as a struct
883 // for backwards compatibility with readers
884 if (convert_ids) {
885 std::int32_t xi, yi;
886 std::uint32_t xu, yu;
887 xi = iptr[0];
888 yi = iptr[1];
889 std::memcpy(&xu, &xi, sizeof(xi));
890 std::memcpy(&yu, &yi, sizeof(yi));
891 ptemp.m_idcpu = ((std::uint64_t)xu) << 32 | yu;
892 } else {
893 ptemp.id() = iptr[0];
894 ptemp.cpu() = iptr[1];
896 iptr += 2;
897
898 for (int j = 0; j < NStructInt; j++)
899 {
900 ptemp.idata(j) = *iptr;
901 ++iptr;
902 }
903
904 AMREX_ASSERT(ptemp.id().is_valid());
905
906 AMREX_D_TERM(ptemp.pos(0) = ParticleReal(rptr[0]);,
907 ptemp.pos(1) = ParticleReal(rptr[1]);,
908 ptemp.pos(2) = ParticleReal(rptr[2]););
909
910 rptr += AMREX_SPACEDIM;
911
912 for (int j = 0; j < NStructReal; j++)
914 ptemp.rdata(j) = ParticleReal(*rptr);
915 ++rptr;
916 }
917
918 locateParticle(ptemp, pld, 0, finestLevel(), IntVect(0));
919
920 std::pair<int, int> ind(grd, pld.m_tile);
921
922 host_real_attribs[lev][ind].resize(NumRealComps());
923 host_int_attribs[lev][ind].resize(NumIntComps());
924
925 // add the struct
926 if constexpr(!ParticleType::is_soa_particle)
927 {
928 host_particles[lev][ind].push_back(ptemp);
929
930 // add the real...
931 for (int icomp = 0; icomp < NumRealComps(); icomp++) {
932 host_real_attribs[lev][ind][icomp].push_back(*rptr);
933 ++rptr;
934 }
935
936 // ... and int array data
937 for (int icomp = 0; icomp < NumIntComps(); icomp++) {
938 host_int_attribs[lev][ind][icomp].push_back(*iptr);
939 ++iptr;
940 }
941 } else {
942 host_particles[lev][ind];
943
944 for (int j = 0; j < AMREX_SPACEDIM; j++) {
945 host_real_attribs[lev][ind][j].push_back(ptemp.pos(j));
946 }
947
948 host_idcpu[lev][ind].push_back(ptemp.m_idcpu);
949
950 // read all other SoA
951 // add the real...
952 for (int icomp = AMREX_SPACEDIM; icomp < NumRealComps(); icomp++) {
953 host_real_attribs[lev][ind][icomp].push_back(*rptr);
954 ++rptr;
955 }
956
957 // ... and int array data
958 for (int icomp = 0; icomp < NumIntComps(); icomp++) {
959 host_int_attribs[lev][ind][icomp].push_back(*iptr);
960 ++iptr;
961 }
962 }
963 }
964
965 for (int host_lev = 0; host_lev < std::ssize(host_particles); ++host_lev)
966 {
967 for (auto& kv : host_particles[host_lev]) {
968 auto grid = kv.first.first;
969 auto tile = kv.first.second;
970 const auto& src_tile = kv.second;
971
972 auto& dst_tile = DefineAndReturnParticleTile(host_lev, grid, tile);
973 auto old_size = dst_tile.size();
974 auto new_size = old_size;
975 if constexpr(!ParticleType::is_soa_particle)
976 {
977 new_size += src_tile.size();
978 } else {
979 amrex::ignore_unused(src_tile);
980 new_size += host_real_attribs[host_lev][std::make_pair(grid,tile)][0].size();
981 }
982 dst_tile.resize(new_size);
983
984 if constexpr(!ParticleType::is_soa_particle)
985 {
986 Gpu::copyAsync(Gpu::hostToDevice, src_tile.begin(), src_tile.end(),
987 dst_tile.GetArrayOfStructs().begin() + old_size);
988 } else {
990 host_idcpu[host_lev][std::make_pair(grid,tile)].begin(),
991 host_idcpu[host_lev][std::make_pair(grid,tile)].end(),
992 dst_tile.GetStructOfArrays().GetIdCPUData().begin() + old_size);
993 }
994
995 for (int i = 0; i < NumRealComps(); ++i) { // NOLINT(readability-misleading-indentation)
997 host_real_attribs[host_lev][std::make_pair(grid,tile)][i].begin(),
998 host_real_attribs[host_lev][std::make_pair(grid,tile)][i].end(),
999 dst_tile.GetStructOfArrays().GetRealData(i).begin() + old_size);
1000 }
1001
1002 for (int i = 0; i < NumIntComps(); ++i) {
1004 host_int_attribs[host_lev][std::make_pair(grid,tile)][i].begin(),
1005 host_int_attribs[host_lev][std::make_pair(grid,tile)][i].end(),
1006 dst_tile.GetStructOfArrays().GetIntData(i).begin() + old_size);
1007 }
1008 }
1009 }
1010
1012}
1013
1014template <typename ParticleType, int NArrayReal, int NArrayInt,
1015 template<class> class Allocator, class CellAssignor>
1016void
1017ParticleContainer_impl<ParticleType, NArrayReal, NArrayInt, Allocator, CellAssignor>
1018::WriteAsciiFile (const std::string& filename)
1019{
1020 BL_PROFILE("ParticleContainer::WriteAsciiFile()");
1021 AMREX_ASSERT(!filename.empty());
1022
1023 const auto strttime = amrex::second();
1024 //
1025 // Count # of valid particles.
1027 Long nparticles = 0;
1029 for (int lev = 0; lev < std::ssize(m_particles); lev++) {
1030 auto& pmap = m_particles[lev];
1031 for (const auto& kv : pmap) {
1032 const auto& aos = kv.second.GetArrayOfStructs();
1033 auto np = aos.numParticles();
1034 Gpu::HostVector<ParticleType> host_aos(np);
1035 Gpu::copy(Gpu::deviceToHost, aos.begin(), aos.begin() + np, host_aos.begin());
1036 for (int k = 0; k < np; ++k) {
1037 const ParticleType& p = host_aos[k];
1038 if (p.id().is_valid()) {
1039 //
1040 // Only count (and checkpoint) valid particles.
1041 //
1042 nparticles++;
1043 }
1044 }
1045 }
1046 }
1047
1048 //
1049 // And send count to I/O processor.
1050 //
1052
1054 {
1055 //
1056 // Have I/O processor open file and write out particle metadata.
1057 //
1058 std::ofstream File;
1059
1060 File.open(filename.c_str(), std::ios::out|std::ios::trunc);
1061
1062 if (!File.good()) {
1063 amrex::FileOpenFailed(filename);
1064 }
1065
1066 File << nparticles << '\n';
1067 File << NStructReal << '\n';
1068 File << NStructInt << '\n';
1069 File << NumRealComps() << '\n';
1070 File << NumIntComps() << '\n';
1071
1072 File.flush();
1073
1074 File.close();
1075
1076 if (!File.good()) {
1077 amrex::Abort("ParticleContainer::WriteAsciiFile(): problem writing file");
1078 }
1079 }
1080
1082
1083 const int MyProc = ParallelDescriptor::MyProc();
1084
1085 for (int proc = 0; proc < ParallelDescriptor::NProcs(); proc++)
1086 {
1087 if (MyProc == proc)
1088 {
1089 //
1090 // Each CPU opens the file for appending and adds its particles.
1091 //
1093
1094 std::ofstream File;
1095
1096 File.rdbuf()->pubsetbuf(io_buffer.dataPtr(), io_buffer.size());
1097
1098 File.open(filename.c_str(), std::ios::out|std::ios::app);
1099
1100 File.precision(15);
1101
1102 if (!File.good()) {
1103 amrex::FileOpenFailed(filename);
1104 }
1105
1106 for (int lev = 0; lev < std::ssize(m_particles); lev++) {
1107 auto& pmap = m_particles[lev];
1108 for (const auto& kv : pmap) {
1109 ParticleTile<ParticleType, NArrayReal, NArrayInt,
1111 pinned_ptile.define(NumRuntimeRealComps(), NumRuntimeIntComps(),
1112 nullptr, nullptr, The_Pinned_Arena());
1113 pinned_ptile.resize(kv.second.numParticles());
1114 amrex::copyParticles(pinned_ptile, kv.second);
1115 const auto& host_aos = pinned_ptile.GetArrayOfStructs();
1116 const auto& host_soa = pinned_ptile.GetStructOfArrays();
1117
1118 auto np = host_aos.numParticles();
1119 for (int index = 0; index < np; ++index) {
1120 const ParticleType* it = &host_aos[index];
1121 if (it->id().is_valid()) {
1122
1123 // write out the particle struct first...
1124 AMREX_D_TERM(File << it->pos(0) << ' ',
1125 << it->pos(1) << ' ',
1126 << it->pos(2) << ' ');
1127
1128 for (int i = 0; i < NStructReal; i++) {
1129 File << it->rdata(i) << ' ';
1130 }
1131
1132 File << it->id() << ' ';
1133 File << it->cpu() << ' ';
1134
1135 for (int i = 0; i < NStructInt; i++) {
1136 File << it->idata(i) << ' ';
1137 }
1138
1139 // then the particle attributes.
1140 for (int i = 0; i < NumRealComps(); i++) {
1141 File << host_soa.GetRealData(i)[index] << ' ';
1142 }
1143
1144 for (int i = 0; i < NumIntComps(); i++) {
1145 File << host_soa.GetIntData(i)[index] << ' ';
1146 }
1147
1148 File << '\n';
1149 }
1150 }
1151 }
1152 }
1153
1154 File.flush();
1155
1156 File.close();
1157
1158 if (!File.good()) {
1159 amrex::Abort("ParticleContainer::WriteAsciiFile(): problem writing file");
1160 }
1161
1162 }
1163
1165 }
1166
1167 if (m_verbose > 1)
1168 {
1169 auto stoptime = amrex::second() - strttime;
1170
1172
1173 amrex::Print() << "ParticleContainer::WriteAsciiFile() time: " << stoptime << '\n';
1174 }
1175}
1176
1177}
1178
1179#endif /*AMREX_PARTICLEIO_H*/
#define BL_PROFILE(a)
Definition AMReX_BLProfiler.H:551
#define AMREX_ASSERT(EX)
Definition AMReX_BLassert.H:38
#define AMREX_ALWAYS_ASSERT(EX)
Definition AMReX_BLassert.H:50
#define AMREX_GPU_HOST_DEVICE
Definition AMReX_GpuQualifiers.H:20
amrex::ParmParse pp
Input file parser instance for the given namespace.
Definition AMReX_HypreIJIface.cpp:15
#define AMREX_D_TERM(a, b, c)
Definition AMReX_SPACE.H:172
Reference-counted collection of Boxes.
Definition AMReX_BoxArray.H:676
Calculates the distribution of FABs to MPI processes.
Definition AMReX_DistributionMapping.H:51
static const IntDescriptor & NativeIntDescriptor()
Returns a constant reference to an IntDescriptor describing the native "int" under which AMReX was co...
Definition AMReX_FPC.cpp:76
Iterator for looping ever tiles and boxes of amrex::FabArray based containers.
Definition AMReX_MFIter.H:88
bool isValid() const noexcept
Is the iterator valid i.e. is it associated with a FAB?
Definition AMReX_MFIter.H:172
A collection (stored as an array) of FArrayBox objects.
Definition AMReX_MultiFab.H:40
const std::string & FileName() const
Definition AMReX_NFiles.H:160
Dynamically allocated vector for trivially copyable data.
Definition AMReX_PODVector.H:308
iterator begin() noexcept
Definition AMReX_PODVector.H:674
Parse Parameters From Command Line and Input Files.
Definition AMReX_ParmParse.H:351
int query(std::string_view name, bool &ref, int ival=FIRST) const
Same as querykth() but searches for the last occurrence of name.
Definition AMReX_ParmParse.cpp:1947
A distributed container for Particles sorted onto the levels, grids, and tiles of a block-structured ...
Definition AMReX_ParticleContainer.H:149
T_ParticleType ParticleType
Definition AMReX_ParticleContainer.H:151
Definition AMReX_GpuAllocators.H:156
This class provides the user with a few print options.
Definition AMReX_Print.H:35
This class is a thin wrapper around std::vector. Unlike vector, Vector::operator[] provides bound che...
Definition AMReX_Vector.H:29
T * dataPtr() noexcept
get access to the underlying data pointer
Definition AMReX_Vector.H:50
Long size() const noexcept
Definition AMReX_Vector.H:54
Vector< Setbuf_Char_Type > IO_Buffer
A simple character buffer for setbuf() usage.
Definition AMReX_VisMFBuffer.H:24
static constexpr int IO_Buffer_Size
We try to do I/O with buffers of this size.
Definition AMReX_VisMFBuffer.H:16
static Long FileOffset(std::ostream &os)
The file offset of the passed ostream.
Definition AMReX_VisMF.cpp:580
static bool GetNoFlushAfterWrite()
Definition AMReX_VisMF.H:284
amrex_particle_real ParticleReal
Floating Point Type for Particles.
Definition AMReX_REAL.H:90
amrex_long Long
Definition AMReX_INT.H:30
bool UseAsyncOut()
Query whether asynchronous output is enabled at runtime.
Definition AMReX_AsyncOut.cpp:70
void WritePlotFile(const std::string &dir, const std::string &name) const
This version of WritePlotFile writes all components and assigns component names.
Definition AMReX_ParticleIO.H:116
Arena * The_Pinned_Arena()
Definition AMReX_Arena.cpp:860
int MyProc() noexcept
Definition AMReX_ParallelDescriptor.H:128
void Barrier(const std::string &)
Definition AMReX_ParallelDescriptor.cpp:1215
void ReduceIntSum(int &)
Definition AMReX_ParallelDescriptor.cpp:1265
void ReadAndBcastFile(const std::string &filename, Vector< char > &charBuf, bool bExitOnError, const MPI_Comm &comm)
Definition AMReX_ParallelDescriptor.cpp:1495
int NProcs() noexcept
Definition AMReX_ParallelDescriptor.H:255
void ReduceLongSum(Long &)
Definition AMReX_ParallelDescriptor.cpp:1236
int IOProcessorNumber() noexcept
The MPI rank number of the I/O Processor (probably rank 0). This rank is usually used to write to std...
Definition AMReX_ParallelDescriptor.H:279
void ReduceLongMax(Long &)
Definition AMReX_ParallelDescriptor.cpp:1237
bool IOProcessor() noexcept
Is this CPU the I/O Processor? To get the rank number, call IOProcessorNumber()
Definition AMReX_ParallelDescriptor.H:289
bool Remove(std::string const &filename)
Remove a file, symbolic link, or empty directory.
Definition AMReX_FileSystem.cpp:41
void copy(HostToDevice, InIter begin, InIter end, OutIter result) noexcept
A host-to-device copy routine. Note this is just a wrapper around memcpy, so it assumes contiguous st...
Definition AMReX_GpuContainers.H:128
void copyAsync(HostToDevice, InIter begin, InIter end, OutIter result) noexcept
A host-to-device copy routine. Note this is just a wrapper around memcpy, so it assumes contiguous st...
Definition AMReX_GpuContainers.H:228
static constexpr DeviceToHost deviceToHost
Definition AMReX_GpuContainers.H:106
static constexpr HostToDevice hostToDevice
Definition AMReX_GpuContainers.H:105
void streamSynchronize() noexcept
Definition AMReX_GpuDevice.H:310
int MyProc()
Definition AMReX_MPMD.cpp:117
void ReduceRealMax(Vector< std::reference_wrapper< Real > > const &)
Definition AMReX_ParallelDescriptor.cpp:1228
Definition AMReX_Amr.cpp:50
void writeIntData(const From *data, std::size_t size, std::ostream &os, const amrex::IntDescriptor &id)
Definition AMReX_IntConv.H:23
__host__ __device__ void ignore_unused(const Ts &...)
No-op helper that marks variables as intentionally unused.
Definition AMReX.H:259
void FileOpenFailed(const std::string &file)
Output a message and abort when couldn't open the file.
Definition AMReX_Utility.cpp:116
void WriteBinaryParticleDataAsync(PC const &pc, const std::string &dir, const std::string &name, const Vector< int > &write_real_comp, const Vector< int > &write_int_comp, const Vector< std::string > &real_comp_names, const Vector< std::string > &int_comp_names, bool is_checkpoint)
Definition AMReX_WriteBinaryParticleData.H:786
void copyParticles(DstTile &dst, const SrcTile &src) noexcept
Copy particles from src to dst. This version copies all the particles, writing them to the beginning ...
Definition AMReX_ParticleTransformation.H:222
__host__ __device__ Dim3 begin(BoxND< dim > const &box) noexcept
Return the iterator begin coordinate of box as Dim3.
Definition AMReX_Box.H:2239
void readFloatData(float *data, std::size_t size, std::istream &is, const RealDescriptor &rd)
Definition AMReX_VectorIO.cpp:136
void writeFloatData(const float *data, std::size_t size, std::ostream &os, const RealDescriptor &rd=FPC::Native32RealDescriptor())
Definition AMReX_VectorIO.cpp:130
bool FileExists(const std::string &filename)
Check if a file already exists. Return true if the filename is an existing file, directory,...
Definition AMReX_Utility.cpp:124
void WriteBinaryParticleDataSync(PC const &pc, const std::string &dir, const std::string &name, const Vector< int > &write_real_comp, const Vector< int > &write_int_comp, const Vector< std::string > &real_comp_names, const Vector< std::string > &int_comp_names, F const &f, bool is_checkpoint)
Definition AMReX_WriteBinaryParticleData.H:486
std::string Concatenate(const std::string &root, int num, int mindigits)
Returns rootNNNN where NNNN == num.
Definition AMReX_String.cpp:35
double second() noexcept
Definition AMReX_Utility.cpp:919
IntVectND< 3 > IntVect
IntVect is an alias for amrex::IntVectND instantiated with AMREX_SPACEDIM.
Definition AMReX_BaseFwd.H:38
void writeDoubleData(const double *data, std::size_t size, std::ostream &os, const RealDescriptor &rd=FPC::Native64RealDescriptor())
Definition AMReX_VectorIO.cpp:142
void Error(const std::string &msg)
Print a message to stderr and abort the program.
Definition AMReX.cpp:235
void readDoubleData(double *data, std::size_t size, std::istream &is, const RealDescriptor &rd)
Definition AMReX_VectorIO.cpp:148
void Abort(const std::string &msg)
Print a fatal-error message to stderr and abort execution.
Definition AMReX.cpp:241
const int[]
Definition AMReX_BLProfiler.cpp:1664
__host__ __device__ Dim3 end(BoxND< dim > const &box) noexcept
Return the iterator end coordinate of box as Dim3.
Definition AMReX_Box.H:2257
void readIntData(To *data, std::size_t size, std::istream &is, const amrex::IntDescriptor &id)
Definition AMReX_IntConv.H:38
Definition AMReX_ParticleIO.H:41
__host__ __device__ int operator()(const P &p) const
Definition AMReX_ParticleIO.H:44
FabArray memory allocation information.
Definition AMReX_FabArray.H:73
MFInfo & SetAlloc(bool a) noexcept
Control whether FAB storage is allocated when the FabArray is defined.
Definition AMReX_FabArray.H:81
uint64_t m_idcpu
Definition AMReX_Particle.H:359
Metadata prefix of an AMReX particle "Header" file.
Definition AMReX_ParticleHeader.H:38
void parse_grid_table(std::istream &is)
Parse the per-level grid table into grids.
Definition AMReX_ParticleHeader.cpp:82
int num_int
number of integer components
Definition AMReX_ParticleHeader.H:52
Vector< Vector< GridEntry > > grids
Per level and grid: where each grid's particle data is stored.
Definition AMReX_ParticleHeader.H:88
Long next_id
the next particle id to hand out (maxnextid)
Definition AMReX_ParticleHeader.H:60
int dim
AMREX_SPACEDIM the file was written with.
Definition AMReX_ParticleHeader.H:46
bool convert_ids
whether particle ids need to be converted (Version_Two_Dot_One and later)
Definition AMReX_ParticleHeader.H:44
void parse(std::istream &is)
Parse the metadata prefix from an input stream.
Definition AMReX_ParticleHeader.cpp:16
int finest_level
finest level present in the file
Definition AMReX_ParticleHeader.H:62
std::string how
precision the particle data was written in: "single" or "double"
Definition AMReX_ParticleHeader.H:42
int num_real
number of real components (pure SoA: excludes the AMREX_SPACEDIM positions)
Definition AMReX_ParticleHeader.H:48
Long num_particles
total number of particles in the file
Definition AMReX_ParticleHeader.H:58
__host__ __device__ bool is_valid() const noexcept
Definition AMReX_Particle.H:252
A struct used for storing a particle's position in the AMR hierarchy.
Definition AMReX_ParticleContainer.H:93
int m_tile
Definition AMReX_ParticleContainer.H:96
The struct used to store particles.
Definition AMReX_Particle.H:405
__host__ __device__ RealVect pos() const &
Definition AMReX_Particle.H:456
__host__ __device__ int & idata(int index) &
Definition AMReX_Particle.H:538
__host__ __device__ ParticleCPUWrapper cpu() &
Definition AMReX_Particle.H:424
__host__ __device__ RealType & rdata(int index) &
Definition AMReX_Particle.H:473
__host__ __device__ ParticleIDWrapper id() &
Definition AMReX_Particle.H:427